/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <math.h>
#include <stdio.h>
#include <GL/glut.h>
#include "draw.h"
#include "drawmenu.h"
#include "drawworld.h"
#include "jomar3D.h"
#include "utils.h"

/* Window infos structure */
struct WindowInfo {
  float ratio; /* ratio of the screen, 0 uninitialized (useful to know when the window has been created) */
  int   mode; /* whether we are in game or windowed mode */
  /* height & width of the window */
  int   width;
  int   height;
};

static struct WindowInfo l_wininfo = {0., 0, 800, 600};


/* set the viewport */
static void
display_setViewport (void)
{
  glViewport (0, 0, l_wininfo.width, l_wininfo.height);
  gluPerspective (80. /* field of view in degrees */, 
                  l_wininfo.ratio, 0.9 /* Z near */, 
                  sqrt ((T_X * T_C) * (T_X * T_C) +
                        (T_Y * T_C) * (T_Y * T_C)) /* Z far */);
}

/* take place in the given window */
void
display_reshape (int width, int height)
{
#if (DEBUG)
  debugmsg (("width, height %d %d\n", width, height));
#endif
  l_wininfo.width = width;
  l_wininfo.height = height;
  l_wininfo.ratio = (float) (width) / (float) (height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  display_setViewport ();
  glMatrixMode (GL_MODELVIEW);
  if (!l_wininfo.mode)
    glutPositionWindow ((glutGet (GLUT_SCREEN_WIDTH) - width) / 2.,
                        (glutGet (GLUT_SCREEN_HEIGHT) - height) / 2.);

/* reset the fps average because it has no reason to stay from the last configuration */
  draw_calculateFPS (1);
  
  drawworld_createMinimap ();
  glutPostRedisplay ();
}

/* set all what must be to launch the game -- graphic part */

/* manage the window: full screen or window */
static void
display_windowManage (void)
{
  char gmsize[20]; /* we hope the resolution is <= to 999999999x999999999... */

/* if it is possible to resize to those crazy values, we cannot resize... then don't */
  sprintf (gmsize, "123x456");
  glutGameModeString (gmsize);
  if (glutGameModeGet (GLUT_GAME_MODE_POSSIBLE))
  {
    sprintf (gmsize, "%dx%d", glutGet (GLUT_SCREEN_WIDTH),
             glutGet (GLUT_SCREEN_HEIGHT));
  }
  else
  {
    sprintf (gmsize, "%dx%d", l_wininfo.width, l_wininfo.height);
  }
  glutGameModeString (gmsize);
  /*
  glutGameModeString( "1600x1200" );
  printf ("Possible Game Mode? %d\n", glutGameModeGet(GLUT_GAME_MODE_POSSIBLE));
  if (!glutGameModeGet(GLUT_GAME_MODE_POSSIBLE)) {
      printf("Resolution of this window not possible for fullscreen\n");
      g_winmode = !g_winmode ;
      return;
  }
  */
  debugmsg (("game mode string %s\n", gmsize));

  if (l_wininfo.mode)
  {
    debugmsg (("Entering game mode\n"));
    
    glutDestroyWindow (glutGetWindow ());
    glutEnterGameMode ();
  }
  else
  {
    debugmsg (("Leaving game mode\n"));
    
    if (l_wininfo.ratio != 0.)
      glutLeaveGameMode ();
    
    glutCreateWindow ("jomar3D");
  }

  glutSetIconTitle ("jomar3D");
  glutSetWindowTitle ("jomar3D");

/* the textures are lost when new gl context so reload them all */
#if (MYFREEGLUT == 0)
  draw_loadTextures (1); /* not if we hold the last context (freeglut_ext), option */
#endif
}

void
display_switchMode (void)
{
  l_wininfo.mode = !l_wininfo.mode;
  display_windowManage ();
#if (MYFREEGLUT == 0)
  draw_setGameAttributes ();        /* only if a new context is used */
  drawmenu_createLists (0);      /* idem */
#endif
}

void
display_setPickport (int x, int y)
{
  GLint viewport[4];
  
  /* fill in the current viewport parameters */
  viewport[0] = 0;
  viewport[1] = 0;
  viewport[2] = l_wininfo.width;
  viewport[3] = l_wininfo.height;

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPickMatrix (x, l_wininfo.height - y, 1.0, 1.0, viewport);
  display_setViewport ();
}

float
display_fitLaps (int pos, int laps, int axis)
{
  if (axis == 1)
    return (float)pos / (float)l_wininfo.width * (float) laps;
  if (axis == 2)
    return (float)pos / (float)l_wininfo.height * (float) laps;
  return 0.;
}

/* Set a default 2D viewport */
void
display_setDefault2D (void)
{
  glLoadIdentity ();
  gluOrtho2D (0, 1200 * l_wininfo.ratio, 0, 1200);
}

void
display_createDisplay (int winuserdef)
{
  glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH |
                       GLUT_MULTISAMPLE);
  
#if (MYFREEGLUT)
  glutSetOption (GLUT_RENDERING_CONTEXT, GLUT_USE_CURRENT_CONTEXT);
  glutSetOption (GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_EXIT);
#endif
  if (!winuserdef)
    glutInitWindowSize (l_wininfo.width, l_wininfo.height);
  else
  {
    l_wininfo.width = glutGet (GLUT_INIT_WINDOW_WIDTH);
    l_wininfo.height = glutGet (GLUT_INIT_WINDOW_HEIGHT);
  }

  display_windowManage ();
#if (MYFREEGLUT)
  draw_loadTextures (1);
#endif
}

void
display_destroyDisplay (void)
{
  if (l_wininfo.mode)
    glutLeaveGameMode ();
}

/* timer called function to hide the cursor */
void
display_hideCursor (int dummy)
{
  (void) dummy;
  glutSetCursor (GLUT_CURSOR_NONE);
}
