/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <GL/glut.h>

#include "draw.h"
#include "drawtext.h"
#include "jomar3D.h"
#include "utf8.h"
#include "utils.h"

#define MYCHARS
#define CHARWIDTH 104.76


/* Functions to call before text writing and after */
void
drawtext_begin (void)
{
  glDisable (GL_DEPTH_TEST);
  glDisable (GL_LIGHTING);
  if (draw_getQuality () != Q_HORRIBLE)
  {
    glEnable (GL_LINE_SMOOTH);
    glEnable (GL_BLEND);
  }
}

void
drawtext_end (void)
{
  glEnable (GL_DEPTH_TEST);
  glEnable (GL_LIGHTING);
  glDisable (GL_LINE_SMOOTH);
  glDisable (GL_BLEND);
}

/** Get the displayed width of a string
 * \param string  an ASCIIZ or null-ended UTF-8 string to print to the screen
 * \return the width needed to display the given string
 * 
 * This function gets the with the given string will take to the screen
 */
float
drawtext_stringWidth (const char *string)
{
#ifdef MYCHARS
  /* we don't use u8_strlen() because it returns -1 if the string have bad
   * characters, but we prefer to display it as bad. */
  size_t len = 0;
  size_t chlen;
  
  while (*string != 0)
  {
    chlen = u8_chlen (string);
    /* bad characters will be displayed, then jump to next byte */
    if (chlen == 0)
      chlen = 1;
    
    len ++;
    
    string += chlen;
  }
  
  return len * CHARWIDTH;
  /*return strlen (string) * CHARWIDTH;*/
#else
  return glutStrokeLength (GLUT_STROKE_ROMAN, (const unsigned char *) string);
#endif
}

/** Draws a string to the screen
 * \param string  an ASCIIZ or null-ended UTF-8 string to print to the screen
 * 
 * This functions draws the given string to the active OpenGL display.
 */
static void
drawtext_drawStringCharacters (const char *string)
{
#define BEGIN glBegin (GL_LINE_STRIP);
#define END   glEnd ();
#define JUMP END BEGIN
#define TOP 119.05
#define MID (TOP / 1.9)
#define BOTTOM -33.33

  const char *p;
  int i;
  float pos, posm, midx;
  long chval = 0;
  size_t chlen;

  for (p = string, i = 0; *p != 0; p += chlen, i++)
  {
    pos  = i * CHARWIDTH + CHARWIDTH / 5.;
    posm = (i + 1) * CHARWIDTH - CHARWIDTH / 5.;
    midx = pos + (posm - pos) / 2.;
    
    chlen = u8_chlen (p);
    /* if the byte is bad, skip it next time */
    if (chlen == 0)
      chlen = 1;
    chval = u8_chval (p);
    
    /* draw the character corresponding to the value */
    BEGIN
    switch (chval)
    {
      case U_a:
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
      break;
      case U_b:
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
      break;
      case U_c:
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_d:
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_e:
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, MID / 2.);
        glVertex2f (pos, MID / 2.);
      break;
      case U_f:
        glVertex2f (midx, BOTTOM);
        glVertex2f (midx, TOP);
        glVertex2f (posm, TOP);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_g:
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, BOTTOM);
        glVertex2f (pos , BOTTOM);
      break;
      case U_h:
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
      break;
      case U_i:
        glVertex2f (midx, 0.);
        glVertex2f (midx, MID);
        JUMP
        glVertex2f (midx, MID * 1.5);
        glVertex2f (midx, TOP);
      break;
      case U_j:
        glVertex2f (pos, 0.);
        glVertex2f (midx, 0.);
        glVertex2f (midx, MID);
        JUMP
        glVertex2f (midx, MID * 1.5);
        glVertex2f (midx, TOP);
      break;
      case U_k:
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (posm, MID);
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, 0.);
      break;
      case U_l:
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (midx, 0.);
      break;
      case U_m:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (midx, 0.);
        glVertex2f (midx, MID);
      break;
      case U_n:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
      break;
      case U_o:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
      break;
      case U_p:
        glVertex2f (pos, BOTTOM);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
      break;
      case U_q:
        glVertex2f (posm, BOTTOM);
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_r:
        glVertex2f (pos + CHARWIDTH / 5., 0.);
        glVertex2f (pos + CHARWIDTH / 5., MID);
        JUMP
        glVertex2f (pos + CHARWIDTH / 5., MID * 0.6);
        glVertex2f (posm - CHARWIDTH / 6., MID);
      break;
      case U_s:
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
      break;
      case U_t:
        glVertex2f (midx, TOP);
        glVertex2f (midx, 0.);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_u:
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
      break;
      case U_v:
        glVertex2f (pos, MID);
        glVertex2f (midx, 0.);
        glVertex2f (posm, MID);
      break;
      case U_w:
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (midx, 0.);
        glVertex2f (midx, MID);
      break;
      case U_x:
        glVertex2f (pos, 0.);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, 0.);
      break;
      case U_y:
        glVertex2f (pos, BOTTOM);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (midx, MID / 2.);
      break;
      case U_z:
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_A:
        glVertex2f (pos, 0.);
        glVertex2f (midx, TOP);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_B:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP * 0.9);
        glVertex2f (posm, TOP * 0.1);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_C:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_D:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP * 0.9);
        glVertex2f (posm, TOP * 0.1);
        glVertex2f (pos, 0.);
      break;
      case U_E:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_F:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_G:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
      break;
      case U_H:
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_I:
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        JUMP
        glVertex2f (midx, TOP);
        glVertex2f (midx, 0.);
        JUMP
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_J:
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        JUMP
        glVertex2f (midx, TOP);
        glVertex2f (midx, 0.);
        glVertex2f (pos, 0.);
      break;
      case U_K:
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (posm, TOP);
        glVertex2f (pos, MID);
        glVertex2f (posm, 0.);
      break;
      case U_L:
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_M:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (midx, MID);
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
      break;
      case U_N:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, 0.);
        glVertex2f (posm, TOP);
      break;
      case U_O:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
      break;
      case U_P:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
      break;
      case U_Q:
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
        glVertex2f (midx, MID);
      break;
      case U_R:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        JUMP
        glVertex2f (midx, MID);
        glVertex2f (posm, 0.);
      break;
      case U_S:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
      break;
      case U_T:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        JUMP
        glVertex2f (midx, TOP);
        glVertex2f (midx, 0.);
      break;
      case U_U:
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, TOP);
      break;
      case U_V:
        glVertex2f (pos, TOP);
        glVertex2f (midx, 0.);
        glVertex2f (posm, TOP);
      break;
      case U_W:
        glVertex2f (pos, TOP);
        glVertex2f (midx, 0.);
        glVertex2f (posm, TOP);
        JUMP
        glVertex2f (midx, 0.);
        glVertex2f (midx, MID);
      break;
      case U_X:
        glVertex2f (pos, TOP);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, 0.);
        glVertex2f (posm, TOP);
      break;
      case U_Y:
        glVertex2f (pos, TOP);
        glVertex2f (midx, MID);
        glVertex2f (posm, TOP);
        JUMP
        glVertex2f (midx, MID);
        glVertex2f (midx, 0.);
      break;
      case U_Z:
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_E_ACUTE:
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (posm, TOP);
      break;
      case U_E_GRAVE:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (pos, TOP);
        glVertex2f (posm, MID * 1.4);
      break;
      case U_e_ACUTE:
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, MID / 2.);
        glVertex2f (pos, MID / 2.);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (posm, TOP);
      break;
      case U_e_CIRCUMFLEX:
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, MID / 2.);
        glVertex2f (pos, MID / 2.);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (midx, TOP);
        glVertex2f (posm, MID * 1.4);
      break;
      case U_A_GRAVE:
        glVertex2f (pos, 0.);
        glVertex2f (midx, MID);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (pos, TOP);
        glVertex2f (posm, MID * 1.4);
      break;
      case U_a_GRAVE:
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (pos, TOP);
        glVertex2f (posm, MID * 1.4);
      break;
      case U_A_ACUTE:
        glVertex2f (pos, 0.);
        glVertex2f (midx, MID);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (posm, TOP);
        glVertex2f (pos, MID * 1.4);
      break;
      case U_a_ACUTE:
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (posm, TOP);
        glVertex2f (pos, MID * 1.4);
      break;
      case U_a_DIAERESIS:
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (pos, TOP);
        glVertex2f (pos, TOP * 0.9);
        JUMP
        glVertex2f (posm, TOP);
        glVertex2f (posm, TOP * 0.9);
      break;
      case U_c_CEDILLA:
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, 0.);
        glVertex2f (posm, BOTTOM / 2.);
        glVertex2f (midx, BOTTOM);
      break;
      case U_o_CIRCUMFLEX:
      case U_O_CIRCUMFLEX:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (midx, TOP);
        glVertex2f (posm, MID * 1.4);
      break;
      case U_o_ACUTE:
      case U_O_ACUTE:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (posm, TOP);
      break;
      case U_o_DIAERESIS:
      case U_O_DIAERESIS:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (pos, TOP);
        glVertex2f (pos, TOP * 0.9);
        JUMP
        glVertex2f (posm, TOP);
        glVertex2f (posm, TOP * 0.9);
      break;
      case U_u_ACUTE:
      case U_U_ACUTE:
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (posm, TOP);
      break;
      case U_u_DIAERESIS:
      case U_U_DIAERESIS:
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (pos, TOP);
        glVertex2f (pos, TOP * 0.9);
        JUMP
        glVertex2f (posm, TOP);
        glVertex2f (posm, TOP * 0.9);
      break;
      case U_n_TILDE:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (midx, TOP);
        glVertex2f (midx, MID * 1.4);
        glVertex2f (posm, TOP);
      break;
      case U_N_TILDE:
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, 0.);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (pos, MID * 1.4);
        glVertex2f (midx, TOP);
        glVertex2f (midx, MID * 1.4);
        glVertex2f (posm, TOP);
      break;
      case U_eszett /* ß */:
        glVertex2f (pos, BOTTOM);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (midx, MID);
        glVertex2f (posm, 0.0);
        glVertex2f (midx, 0.);
      break;
      case U_0:
        glVertex2f (pos, 0.);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (posm, TOP);
      break;
      case U_1:
        glVertex2f (pos, MID);
        glVertex2f (midx, TOP);
        glVertex2f (midx, 0.);
        JUMP
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_2:
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, MID);
        glVertex2f (pos, MID);
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
      break;
      case U_3:
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_4:
        glVertex2f (pos, TOP);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (midx, TOP);
        glVertex2f (midx, 0.);
      break;
      case U_5:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
      break;
      case U_6:
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
      break;
      case U_7:
        glVertex2f (pos, TOP);
        glVertex2f (midx, TOP);
        glVertex2f (midx, 0.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_8:
        glVertex2f (pos, MID);
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_9:
        glVertex2f (pos, 0.);
        glVertex2f (posm, 0.);
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (pos, MID);
        glVertex2f (posm, MID);
      break;
      case U_SPACE:;
      break;
      case U_UNDERSCORE /* '_' */:
        glVertex2f (pos, BOTTOM);
        glVertex2f (posm, BOTTOM);
      break;
      case U_FULL_STOP /* '.' */:
        glVertex2f (midx, 0.);
        glVertex2f (midx, TOP * 0.1);
      break;
      case U_COLON /* ':' */:
        glVertex2f (midx, 0.);
        glVertex2f (midx, TOP * 0.1);
        JUMP
        glVertex2f (midx, MID);
        glVertex2f (midx, MID * 0.8);
      break;
      case U_SEMICOLON /* ';' */:
        glVertex2f (midx, 0.);
        glVertex2f (pos, BOTTOM);
        JUMP
        glVertex2f (midx, MID);
        glVertex2f (midx, MID * 0.8);
      break;
      case U_COMMA /* ',' */:
        glVertex2f (midx, 0.);
        glVertex2f (pos, BOTTOM);
      break;
      case U_LEFT_PARENTHESE /* '(' */:
        glVertex2f (midx, TOP);
        glVertex2f (pos, MID);
        glVertex2f (midx, 0.);
      break;
      case U_RIGHT_PARENTHESE /* ')' */:
        glVertex2f (midx, TOP);
        glVertex2f (posm, MID);
        glVertex2f (midx, 0.);
      break;
      case U_SLASH /* '/' */:
        glVertex2f (posm, TOP);
        glVertex2f (pos, 0.);
      break;
      case U_APOSTROPHE /* '\'' */:
        glVertex2f (midx, TOP);
        glVertex2f (midx, MID);
      break;
      case U_HYPEN /* '-' */:
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
      break;
      case U_PLUS /* '+' */:
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (midx, MID);
        glVertex2f (midx, 0.);
      break;
      case U_ASTERISK /* '*' */:
        glVertex2f (pos, 0.);
        glVertex2f (posm, MID);
        JUMP
        glVertex2f (pos, MID / 2.);
        glVertex2f (posm, MID / 2.);
        JUMP
        glVertex2f (pos, MID);
        glVertex2f (posm, 0.);
      break;
      case U_AMPERSAND /* '&' */:
        glVertex2f (posm, MID);
        glVertex2f (midx, 0.);
        glVertex2f (pos, 0.);
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
        glVertex2f (posm, 0.);
      break;
      case U_EXCLAMATION_MARK /* '!' */:
        glVertex2f (midx, TOP);
        glVertex2f (midx, MID / 2.);
        JUMP
        glVertex2f (midx, 0.);
        glVertex2f (midx, TOP * 0.1);
      break;
      case U_QUESTION_MARK /* '?' */:
        glVertex2f (pos, TOP);
        glVertex2f (posm, TOP);
        glVertex2f (posm, MID);
        glVertex2f (midx, MID);
        glVertex2f (midx, MID / 2.);
        JUMP
        glVertex2f (midx, 0.);
        glVertex2f (midx, TOP * 0.1);
      break;
      /* inavlid or unimplemented character */
      default:
        glVertex2f (pos, TOP);
        glVertex2f (posm, BOTTOM);
        glVertex2f (pos, BOTTOM);
        glVertex2f (posm, TOP);
        glVertex2f (pos, TOP);
      break;
    }
    END
  }
#undef BOTTOM
#undef MID
#undef TOP
#undef JUMP
#undef END
#undef BEGIN
}

/** Draws a text on the window
 * @param x, y            Coordinate reference for the text
 * @param factx, facty    Multiplication factor of the normal size of the text
 * @param align           Alignment of the text from its coorindates
 * @param text_max_width  Maximum width of the text
 * @param text_maximize   Whether the text be miximized (be text_max_width units long)
 * @param draw_bg         Whether a background should be drawn behind the text
 * @param text            The pointer to the text to be drawn
 */
/* very useful routine to write text to the OpenGL window including scaling and background drawing */
void
drawtext_drawString (GLfloat x, GLfloat y, GLfloat factx, GLfloat facty,
                     int align, float text_max_width, int text_maximize, int draw_bg,
                     const char *text)
{
  GLfloat width;
  width = drawtext_stringWidth (text);
#if (DEBUG == 2)
  debugmsg (("textwidth %f\n", width * factx));
#endif
  
  glPushMatrix ();
  glTranslatef (x, y, 0);
  if (width * factx > text_max_width || text_maximize)
    factx = text_max_width / width;
  glScalef (factx, facty, 1.);
  if (align == TEXT_CENTER)
    glTranslatef (-width / 2., 0, 0);
  if (align == TEXT_RIGHT)
    glTranslatef (-width, 0, 0);

/* draw a background for the text so it is much better readable (when flag on) */
  if (draw_bg)
  {
    GLfloat textColor[4];
    
    glGetFloatv (GL_CURRENT_COLOR, textColor);
    glColor4f (1. - textColor[0], 1. - textColor[1], 1. - textColor[2],
               0.6);
    glRecti (-20, -40, width + 20, 130);
    glColor4fv (textColor);
  }

#ifdef MYCHARS
  drawtext_drawStringCharacters (text);
#else
/* print the characters -- take care of some special character significations */
  for (; *text; text++)
  {
    if (*text == '`' || *text == '°' || *text == '^' || *text == '~' || *text == '"')
    {
      glPushMatrix ();
      glTranslatef (0.0, 55., 0.0);
      if (*text == '`')
        glTranslatef (0., -35.0, 0.);
      if (*text == '^')
        glScalef (0.7, 0.7, 1.);
      if (*text == '"')
        glTranslatef (12., -30.0, 0.);
      if (*text == '°')
      {
        glRotatef (90., 0, 0, 1);
        glTranslatef (0., -120.0, 0.);
        glutStrokeCharacter (GLUT_STROKE_ROMAN, '`');
      }
      else
      {
        glutStrokeCharacter (GLUT_STROKE_ROMAN, *text);
      }
      glPopMatrix ();
    }
    else
    {
      glutStrokeCharacter (GLUT_STROKE_ROMAN, *text);
    }
  }
#endif /* MYCHARS */
  glPopMatrix ();
}
