/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "jomar3D.h"
#if (USEGLEW)
#  include <GL/glew.h>
#endif
#include <GL/glut.h>

#include "core.h"
#include "draw.h"
#include "drawworld.h"
#include "utils.h"


static int l_wells[WALLLONG][NEWALL][3]; /* 1st: size of the wall
                                            2de: wall number
                                            3de: 0 x position, 1 y position, 2 direction */

static GLfloat *l_wellsdata = NULL;      /* pointer to all wells vertices */

static int l_vertwall1 = 8;       /* number of vertices of wall 1 */
static int l_verteachwall = 18;   /* number of vertices of each wall other than wall 1 */
#define VERTICETOT ((l_vertwall1 + l_verteachwall * (WALLLONG - 1)) * 5)
static int l_ndivwells = 1;       /* number of times to divide a quarter of circle of the wells' end */

/* define all colors */
static GLfloat l_pinceColor[]   = { 1.0, 0.3, 0.3, 1.0 };
static GLfloat l_armColor[]     = { 1.0, 0.5, 0.5, 1.0 };
static GLfloat l_tailColor[]    = { 0.7, 0.2, 0.2, 1.0 };
static GLfloat l_bodyColor[]    = { 0.8, 0.4, 0.4, 1.0 };
static GLfloat l_eyesColor[]    = { 0.2, 0.2, 0.2, 1.0 };
static GLfloat l_monthColor[]   = { 0.0, 0.5, 0.5, 1.0 };
static GLfloat l_legsColor[]    = { 1.0, 0.5, 0.2, 1.0 };
static GLfloat l_mskinColor[]   = { 0.2, 0.9, 0.3, 1.0 };
static GLfloat l_meyeColor[]    = { 1.0, 0.8, 0.8, 1.0 };
static GLfloat l_soccerColor[]  = { 0.9, 0.3, 0.3, 1.0 };
static GLfloat l_glassColor[]   = { 0.4, 0.4, 1.0, 1.0 };
static GLfloat l_boosterColor[] = { 1.0, 0.8, 0.3, 1.0 };
static GLfloat l_pupilleColor[] = { 0.0, 0.0, 0.3, 1.0 };
static GLfloat l_fireColor[]    = { 1.0, 1.0, 0.0, 0.5 };
static GLfloat l_ceilingColor[] = { 0.6, 0.6, 1.0, 1.0 };
/*static GLfloat l_lightColor[]   = { 1.0, 1.0, 1.0, 1.0 };*/
/*static GLfloat l_earthColor[]   = {0.0, 1.0, 0.0, 1.0};*/
/*static GLfloat l_antColor[]     = {0.0, 0.0, 1.0, 0.5};*/
/*static GLfloat l_fogColor[]     = {0.8, 0.8, 0.9, 1.0};*/


#if USEGLEW && defined (GL_ARB_vertex_buffer_object)
static GLuint l_vbobuf = 0;       /* vbo pointer */
#endif


/* Funtion to store a wall to draw, called at init */
void
drawworld_rememberWall (int length, int number, int x, int y, int dir)
{
  l_wells[length][number][0] = x;
  l_wells[length][number][1] = y;
  l_wells[length][number][2] = dir;
  
  debugmsg (("wall position: %d %d length, number %d %d\n",
            x, y, length, number));
}

/* Notice that the number of dividers of the wells has changed */
void
drawworld_setWellsDividers (int n)
{
  l_ndivwells = n;
}

/* XXXXXXXXXXXXXXXXX INITIALIZATION AND DRAWING OF THE LOBSTER XXXXXXXXXXXXXXXXXXXXXX */
/* create drawing lists for the lobster */

static void
drawworld_makeLobsterPincer (void)
{
  glNewList (PINCER, GL_COMPILE);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_pinceColor);
  gluCylinder (draw_getQuadric (), 6.0, 9.0, 16.0, 7*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_INSIDE);
  gluDisk (draw_getQuadric (), 0.0, 6.0, 7*draw_getVertMult (), 1);
  glTranslatef (0.0, 0.0, 16.0);
  gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE);
  gluDisk (draw_getQuadric (), 0.0, 9.0, 7*draw_getVertMult (), 1);
  glTranslatef (4.5, 0.0, -1.0);
  glRotatef (-15, 0.0, 1.0, 0.0);
  gluCylinder (draw_getQuadric (), 4.0, 1.0, 25.0, 5*draw_getVertMult (), 1);
  glRotatef (15, 0.0, 1.0, 0.0);
  glTranslatef (-9.9, 0.0, 1.0);
  gluCylinder (draw_getQuadric (), 3.0, 0.0, 20.0, 5*draw_getVertMult (), 1);
  glEndList ();
}

static void
drawworld_makeLobsterArms (void)
{
  glNewList (ARM, GL_COMPILE);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_armColor);
  glRotatef (28, 0.0, 1.0, 0.0);
  gluCylinder (draw_getQuadric (), 7., 5., 15.0, 5*draw_getVertMult (), 1);
  glTranslatef (0.0, 0.0, 15.0);
  gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE);
  gluDisk (draw_getQuadric (), 0.0, 5., 10, 1);
  glTranslatef (0.0, 0.0, -1.8);
  glRotatef (-30, 0.0, 1.0, 0.0);
  gluCylinder (draw_getQuadric (), 4.0, 3.5, 14.0, 5*draw_getVertMult (), 1);
  glTranslatef (0.0, 0.0, 14.0);
  glEndList ();
}

static void
drawworld_drawLobsterArms (void)
{
  glRotatef (10., 0.0, 1.0, 0.0);
  glCallList (ARM);

  glTranslatef (0.0, 0.0, -0.6);
  glScalef (1.0, 0.6, 1.0);
  glRotatef (-10. + core_getLobster ()->phase / 2., 0.0, 1.0, 0.0);
  glCallList (PINCER);
}

static void
drawworld_makeLobsterAnt (void)
{
  int i;

  glNewList (ANT, GL_COMPILE);
  glLineWidth (2);
  glBegin (GL_LINE_STRIP);
  for (i = 0; i < 6; i++)
    glVertex3f (sqrt (i*2), 0, -i);
  glVertex3f (9., 0., -60.);
  glEnd ();
  glEndList ();
}

static void
drawworld_makeLobsterLegs (void)
{
  glNewList (LEG, GL_COMPILE);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_legsColor);
  glRotatef (90, 1.0, 0.0, 0.0);
  glRotatef (50, 0.0, 1.0, 0.0);
  gluCylinder (draw_getQuadric (), 2., 0.5, 15.0, 3*draw_getVertMult (), 1);
  glTranslatef (0.0, 0.0, 14.9);
  glRotatef (-20, 0.0, 1.0, 0.0);
  gluQuadricOrientation (draw_getQuadric (), GLU_INSIDE);
  gluDisk (draw_getQuadric (), 0.0, 1.6, 3*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE);
  gluCylinder (draw_getQuadric (), 1.6, 0.5, 13.0, 3*draw_getVertMult (), 1);
  glTranslatef (0.0, 0.0, 12.9);
  glRotatef (-10, 0.0, 1.0, 0.0);
  gluQuadricOrientation (draw_getQuadric (), GLU_INSIDE);
  gluDisk (draw_getQuadric (), 0.0, 1.4, 3*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE);
  gluCylinder (draw_getQuadric (), 1.4, 0.2, 6.0, 3*draw_getVertMult (), 1);
  glEndList ();
}

void
drawworld_makeLobsterTail (void)
{
  int boostflag = core_getBoostState ();
  glNewList (TAIL, GL_COMPILE);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_tailColor);
  if (boostflag == 0)
  {
    gluSphere (draw_getQuadric (), 10.0, 10*draw_getVertMult (), 5*draw_getVertMult ());
    glRotatef (-5, 1.0, 0.0, 0.0);
    glRotatef (10, 0.0, 1.0, 0.0);
    glTranslatef (0, 0.0, -10.0);
  }
  if (boostflag < 2)
  {
    gluSphere (draw_getQuadric (), 8.0, 9*draw_getVertMult (), 5*draw_getVertMult ());
    glRotatef (-5, 1.0, 0.0, 0.0);
    glRotatef (10, 0.0, 1.0, 0.0);
    glTranslatef (0, 0.0, -8.0);
  }
  if (boostflag < 3)
  {
    gluSphere (draw_getQuadric (), 7.0, 7*draw_getVertMult (), 5*draw_getVertMult ());
    glTranslatef (-5.0, 0.0, -7.0);
    glRotatef (90, 1.0, 0.0, 0.0);
    glRotatef (10, 0.0, 1.0, 0.0);
  }
  if (boostflag == 3)
  {
    glTranslatef (00.0, 3.0, 00.0);
    gluSphere (draw_getQuadric (), 3.0, 5*draw_getVertMult (), 5*draw_getVertMult ());
    glTranslatef (-5.0, 0.0, -3.0);
    glRotatef (90, 1.0, 0.0, 0.0);
    glRotatef (10, 0.0, 1.0, 0.0);
  }

  gluQuadricOrientation (draw_getQuadric (), GLU_INSIDE);
  gluDisk (draw_getQuadric (), 0.0, 8.0, 10*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE ) ;
  gluDisk (draw_getQuadric (), 0.0,  8.0, 10*draw_getVertMult (), 1);
  glTranslatef (5.0, 0.0, 1.0);
  glRotatef (-10, 1.0, 1.0, 0.0);
  gluDisk (draw_getQuadric (), 0.0, 7.0, 7*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_INSIDE ) ;
  gluDisk (draw_getQuadric (), 0.0,  7.0, 7*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE ) ;
  glTranslatef (5.0, 2.0, 1.0);
  glRotatef (-10, 1.0, 1.0, 0.0);
  gluDisk (draw_getQuadric (), 0.0, 6.0, 7*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_INSIDE ) ;
  gluDisk (draw_getQuadric (), 0.0,  6.0, 7*draw_getVertMult (), 1);
  gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE);
  glEndList ();
}

static void
drawworld_makeLobsterBody (void)
{
  glNewList (BODY, GL_COMPILE);
  glTranslatef (0.0, 6.0, 0.0);
  /* eyes */
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_eyesColor);
  glTranslatef (3.5, 4.0, 0.0);
  gluSphere (draw_getQuadric (), 2.0, 5*draw_getVertMult (), 5*draw_getVertMult ());
  glTranslatef (-7.0, 0.0, 0.0);
  gluSphere (draw_getQuadric (), 2.0, 5*draw_getVertMult (), 5*draw_getVertMult ());
  glTranslatef (3.5, -4.0, 0.1);
  /* mouth never seen */
  /*
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_monthColor);
  luPartialDisk (draw_getQuadric (), 0.0, 3.0, 15, 1, 45., 270.);
  glTranslatef (0.0, 0.0, -0.1);
  */
  /* body */
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_pinceColor);
  gluDisk (draw_getQuadric (), 0.0, 5.0, 7*draw_getVertMult (), 1);
  glTranslatef (0.0, 0.0, -7.0);
  gluCylinder (draw_getQuadric (), 6.0, 5., 7.0, 7*draw_getVertMult (), 1);
  glRotatef (-10, 1.0, 0.0, 0.0);
  glTranslatef (0.0, 0.0, 3.7);
  glRotatef (-20, 1.0, 0.0, 0.0);
  gluDisk (draw_getQuadric (), 0.0, 7.0, 7*draw_getVertMult (), 1);
  glTranslatef (0.0, 0.0, -16.0);
  gluCylinder (draw_getQuadric (), 11.8, 7., 16.0, 7*draw_getVertMult (), 1);
  glRotatef (30, 1.0, 0.0, 0.0);
  glTranslatef (0.0, 0.0, -16.0);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_bodyColor);
  /*gluSphere (draw_getQuadric (), 12.0, 20, 10);*/
  glTranslatef (0.0, 0.0, -12.0);
  /*gluSphere (draw_getQuadric (), 11.0, 20, 10);*/
  /*glTranslatef (0.0, 0.0, -11.0);*/
  gluCylinder (draw_getQuadric (), 10., 10.7, 33.0, 7*draw_getVertMult (), 1);
  gluSphere (draw_getQuadric (), 11.0, 10*draw_getVertMult (), 5*draw_getVertMult ());
  glEndList ();
}

/* draw the lobster using the lists below */
void
drawworld_drawLobster (void)
{
  Lobster *lobster = core_getLobster ();
  
  glTranslatef (0, 0, -29);
  glPushMatrix ();
  glTranslatef (04, 0, 0);
  glRotatef (lobster->phase / 2, 0.0, 1.0, 1.0);
  drawworld_drawLobsterArms ();
  glPopMatrix ();
  glPushMatrix ();
  glTranslatef (-04, 0, 0);
  glRotatef (lobster->phase / 2, 0.0, 1.0, 1.0);
  glRotatef (180., 0.0, 0.0, 1.0);
  drawworld_drawLobsterArms ();
  glPopMatrix ();

  glPushMatrix ();
  glTranslatef (0, 0, 15);
  glCallList (BODY);
  glTranslatef (0, 0, -12);
  glRotatef (lobster->phase, 0.0, 1.0, 1.0);
  glCallList (TAIL);
  glPopMatrix ();

  glPushMatrix ();
  glTranslatef (5, 7, 14);
  glRotatef (lobster->phase / 1.7, 1.0, 1.0, 1.0);
  glCallList (ANT);
  glPopMatrix ();
  glPushMatrix ();
  glTranslatef (-5, 7, 14);
  glRotatef (180., 0.0, 0.0, 1.0);
  glRotatef (lobster->phase / 1.7, 1.0, 1.0, 1.0);
  glCallList (ANT);
  glPopMatrix ();

  glPushMatrix ();
  glTranslatef (8.0, 0, -6);
  glPushMatrix ();
  glRotatef (lobster->legangle[0], 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glTranslatef (0.0, 0.0, -8);
  glPushMatrix ();
  glRotatef (lobster->legangle[1], 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glTranslatef (0.0, 0.0, -8);
  glPushMatrix ();
  glRotatef (lobster->legangle[2], 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glTranslatef (0.0, 0.0, -8);
  glPushMatrix ();
  glRotatef (lobster->legangle[3], 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glPopMatrix ();

  glPushMatrix ();
  glTranslatef (-8.0, 0, -6);
  glRotatef (180., 0.0, 1.0, 0.0);
  glPushMatrix ();
  glRotatef (-(lobster->legangle[0]), 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glTranslatef (0.0, 0.0, 8);
  glPushMatrix ();
  glRotatef (-(lobster->legangle[1]), 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glTranslatef (0.0, 0.0, 8);
  glPushMatrix ();
  glRotatef (-(lobster->legangle[2]), 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glTranslatef (0.0, 0.0, 8);
  glPushMatrix ();
  glRotatef (-(lobster->legangle[3]), 0.0, 1.0, 0.0);
  glCallList (LEG);
  glPopMatrix ();
  glPopMatrix ();
}

/* XXXXXXXXXXXXXXXXX INITIALIZATION AND DRAWING OF THE MARTIAN  XXXXXXXXXXXXXXXXXXXXXX */
/* create the lists for the Martian */

static void
drawworld_makeMartianHead (void)
{
  glNewList (HEAD, GL_COMPILE);

  /* body and head */
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_mskinColor);
  glPushMatrix ();
    glTranslatef (0.0, 5.0, 3.0);
    glScalef (1.0, 0.6, 0.6);
    gluSphere (draw_getQuadric (), 12.0, 5*draw_getVertMult (), 5*draw_getVertMult ());
    glTranslatef (0.0, 20.0, 0.0);
    gluSphere (draw_getQuadric (), 10.0, 5*draw_getVertMult (), 5*draw_getVertMult ());
  glPopMatrix ();

  /* white of the eye */
  glMaterialfv (GL_FRONT, GL_DIFFUSE, l_meyeColor);
  glPushMatrix ();
    glTranslatef (0.0, 4.5 + 17., -2.5 + 3.);
    gluSphere (draw_getQuadric (), 2.5, 3*draw_getVertMult (), 3*draw_getVertMult ());
    glTranslatef (-5., -1.2, 1.);
    gluSphere (draw_getQuadric (), 2.5, 3*draw_getVertMult (), 3*draw_getVertMult ());
    glTranslatef (10., 0., 0.);
    gluSphere (draw_getQuadric (), 2.5, 3*draw_getVertMult (), 3*draw_getVertMult ());
  glPopMatrix ();

  /* black of the eye */
  if (draw_getQuality () == Q_FINE)
  {
    glMaterialfv (GL_FRONT, GL_DIFFUSE, l_pupilleColor);
    glPushMatrix ();
      glTranslatef (0.0, 5.5 + 17., -4.5 + 3.);
      gluSphere (draw_getQuadric (), 0.5, 4, 2);
      glTranslatef (-5., -1.2, 1.);
      gluSphere (draw_getQuadric (), 0.5, 4, 2);
      glTranslatef (10., 0., 0.);
      gluSphere (draw_getQuadric (), 0.5, 4, 2);
    glPopMatrix ();
  }

  /* antennae */
  glPushMatrix ();
    glTranslatef (0.0, 5. + 17., 2.0 + 3.);
    glRotatef (-90., 1.0, 0.0, 0.0);
    glRotatef (20., 0.0, 1.0, 0.0);
    glMaterialfv (GL_FRONT, GL_DIFFUSE, l_mskinColor);
    gluCylinder (draw_getQuadric (), 0.5, 0.0, 7.0, 2*draw_getVertMult (), 1);
    glRotatef (-40., 0.0, 1.0, 0.0);
    gluCylinder (draw_getQuadric (), 0.5, 0.0, 7.0, 2*draw_getVertMult (), 1);
    glTranslatef (0.0, 0.0, 7.0);
    glMaterialfv (GL_FRONT, GL_DIFFUSE, l_soccerColor);
    gluSphere (draw_getQuadric (), 1.0, 2*draw_getVertMult (), 2*draw_getVertMult ());
    glRotatef (20., 0.0, 1.0, 0.0);
    glTranslatef (5.0, 0.0, 0.0);
    gluSphere (draw_getQuadric (), 1.0, 2*draw_getVertMult (), 2*draw_getVertMult ());
  glPopMatrix ();

  /* mouth */
  if (draw_getQuality () == Q_FINE)
  {
    glPushMatrix ();
      glTranslatef (0.0, 1. + 17., -6. + 3.);
      glRotatef (25., 1.0, 0.0, 0.0);
      glScalef (3.0, 1.0, 1.0);
      glMaterialfv (GL_FRONT, GL_DIFFUSE, l_monthColor);
      gluQuadricOrientation (draw_getQuadric (), GLU_INSIDE);
      gluPartialDisk (draw_getQuadric (), 0.0, 1.5, 10, 1, 55., 260.);
      gluQuadricOrientation (draw_getQuadric (), GLU_OUTSIDE);
    glPopMatrix ();
  }

  glEndList ();
}

static void
drawworld_makeMartianSoccer (void)
{
#define ndot 200
#define angle (M_PI / 12.)
#define rayon 20.
  int i;
  float randa, randr, randp, posz;

  glNewList (SOCCER, GL_COMPILE);

  /* soccer */
  glPushMatrix ();
    glMaterialfv (GL_FRONT, GL_DIFFUSE, l_soccerColor);
    glScalef (1.0, 0.4, 1.5);
    gluSphere (draw_getQuadric (), 15., 4*draw_getVertMult (), 6*draw_getVertMult ());
  glPopMatrix ();

  /* boosters */
  glPushMatrix ();
    glTranslatef (4.5, 2.0, 18.0);
    glScalef (1.0, 0.6, 1.0);
    glDisable(GL_CULL_FACE);
    glMaterialfv (GL_FRONT, GL_DIFFUSE, l_boosterColor);
    gluCylinder (draw_getQuadric (), 2., 3., 8., 3*draw_getVertMult (), 1);
    glTranslatef (-9.0, 0.0, 0.0);
    gluCylinder (draw_getQuadric (), 2., 3., 8., 3*draw_getVertMult (), 1);
    glEnable(GL_CULL_FACE);
  glPopMatrix ();

  /* glass */
  glPushMatrix ();
    glTranslatef (0.0, 4.0, -4.0);
    glRotatef (-90., 1.0, 0.0, 0.0);
    glScalef (1.0, 1.4, 1.0);
    glMaterialfv (GL_FRONT, GL_DIFFUSE, l_glassColor);
    glDisable(GL_CULL_FACE);
    gluCylinder (draw_getQuadric (), 10.0, 8.0, 5., 5*draw_getVertMult (), 1);
    glEnable(GL_CULL_FACE);
  glPopMatrix ();

  glEndList ();

  glNewList (FIRE, GL_COMPILE);

  glPushMatrix ();
    glTranslatef (4.5, 2.0, 20.0);
    glDisable (GL_LIGHTING);
    glColor4f (l_fireColor[0], l_fireColor[1], l_fireColor[2], l_fireColor[3]);
    glBegin (GL_POINTS);
      for (i = 0; i < ndot; i++)
      {
        randa = (MYRANDSYM * 2.) * angle;
        randr = MYRAND * rayon;
        posz = sin (randa) * randr;
        randp = MYRANDSYM * 2. *
                sqrt (pow (randr * sin (angle), 2) - posz * posz);
        glVertex3f (randp, posz, cos (randa) * randr);
        glVertex3f (randp-9., posz, cos (randa) * randr);
      }
    glEnd ();
    glEnable (GL_LIGHTING);
  glPopMatrix ();

  glEndList ();
#undef ndot
#undef angle
#undef rayon
}

/* draw the Martian using the lists below */
void
drawworld_drawMartian (int state)
{
  /* whether we are rendering the shadow */
  if (state & SHADOW)
  {
    glRotatef (-90., 1, 0, 0);
    glScalef (1., 2., 1.);
    gluDisk (draw_getQuadric (), 0.0, 15., 5*draw_getVertMult (), 1);
  }
  else
  {                           
    glCallList (SOCCER);
    glCallList (HEAD);

    /* boosters are on ? */
    if (state & ON)
    {
      glRotatef (MYRANDSYM*3., 1, 1, 1);
      glCallList (FIRE);
    }
  }
}


/* XXXXXXXXXXXXXXXXX DRAWING OF THE BOMB  XXXXXXXXXXXXXXXXXXXXXX */

/* Draw the burning bomb */
void
drawworld_drawBurningBomb (Bomb* bomb)
{
  int i;
  float posv, percent;

  percent = (bomb->duration - (glutGet(GLUT_ELAPSED_TIME) - bomb->throwtime)) /
            (float) bomb->duration;


  glPushMatrix ();

  /* Shadow */
  glEnable (GL_BLEND);
  glTranslatef (-(bomb->pos.x) * T_C, 1., bomb->pos.y * T_C);
  glRotatef (-90., 1, 0, 0);
  glColor4f (0., 0., 0., 0.3);
  gluDisk (draw_getQuadric (), 0.0, 8., 5*draw_getVertMult (), 1);
  glRotatef (90., 1, 0, 0);
  glDisable (GL_BLEND);
  /* Sphere */
  glTranslatef (0., 7., 0.);
  glColor3f (0., 0., 0.);
  glutSolidSphere (8., 15, 7);
  /* Wick */
  glRotatef (bomb->angle.x, 0, 0, 1);
  glRotatef (bomb->angle.z, 1, 0, 0);
  glTranslatef (0., 8., 0.);
  glColor3f (.8, .8, .8);
  glLineWidth (3.);
  glBegin (GL_LINE_STRIP);
    for (i = 0; i < 60 * percent; i++)
      glVertex3f (2 * cos (i/2.), i/2., 2 * sin (i/2.));
  glEnd ();
  /* Smoke & flames */
  glTranslatef (2 * cos (60*percent/2.), (60*percent/2.), (2 * sin (60*percent/2.)));
  glRotatef (-(bomb->angle.x), 0, 0, 1);
  glRotatef (-(bomb->angle.z), 1, 0, 0);
  glColor3f (.9, .9, .5);
  glBegin (GL_POINTS);
    for (i = 0; i < 400; i++)
    {
      posv = MYRAND * 10. + 3.;
      glVertex3f ((MYRANDSYM) * pow (posv/4., 2.),
                  posv - 3.,
                  (MYRANDSYM) * pow (posv/4., 2.));
    }
  glEnd ();

  glPopMatrix ();

}

/* Draw the exploding bomb */
void
drawworld_drawExplodingBomb (Bomb* bomb)
{
  int i;
  float percent, fact, angle1, angle2;
#define SIZEMAX T_C * 2.
#define FRAC 16.        /* Number with which I divide the EXPLODE time to have the growing time */
  percent = 1. - (EXPLODE/FRAC - (glutGet(GLUT_ELAPSED_TIME) - bomb->throwtime - bomb->duration)) / EXPLODE * FRAC;
  fact = percent;
  if (percent > 1.)
    fact = 1. - percent / FRAC;
  if (percent > 1.)
    percent = 1.;
#undef FRAC
  
  glPushMatrix ();
  glTranslatef (-(bomb->pos.x) * T_C, 1., bomb->pos.y * T_C);
  glColor3f (.8 + (MYRANDSYM), .8, 0.);
  glBegin (GL_POINTS);
    for (i = 0; i < 20000 * fact; i++)
    {
      angle1 = MYRAND*M_PI*2.;
      angle2 = MYRAND*M_PI;
      glVertex3f (cos(angle1)*cos(angle2)*SIZEMAX*percent,
                  sin(angle2)*SIZEMAX*percent,
                  sin(angle1)*cos(angle2)*SIZEMAX*percent);
    }
  glEnd ();
#undef SIZEMAX

  glColor3f (1., 0., 0.);
  glBegin (GL_POINTS);
    for (i = 0; i < 200 * fact; i++)
    {
      angle1 = MYRAND*M_PI*2.;
      glVertex3f (MYRAND * cos(angle1) * T_C * percent / 4.,
                  MYRAND * 20. * fact,
                  MYRAND * sin(angle1) * T_C * percent / 4.);
    }
  glEnd ();
  glEnable(GL_BLEND);
  glColor4f (0., 0., 0., fact / 1.5);
  glRotatef (-90, 1, 0, 0);
  gluDisk (draw_getQuadric (),  0., T_C * percent / 2., 5*draw_getVertMult (), 1);
  glDisable(GL_BLEND);

  glPopMatrix ();
  
}

/* Clear the wells vertices array and vbo */
void
drawworld_clearWells (void)
{
  if (l_wellsdata)
  {
    free (l_wellsdata);
#if USEGLEW && defined (GL_ARB_vertex_buffer_object)
    if (GLEW_ARB_vertex_buffer_object)
      glDeleteBuffersARB (1, &l_vbobuf);
#endif
  }
}

void
drawworld_makeWellsData (void)
{
  int i = 1;
  int j = 0;
/* be a little smaller so that when the wall is next to another it doesn't flicker */
#define HALF (T_C / 2. - 0.1)
/* That is the value to give to draw a unit of texture otherwise is may repeat like 1.1 */
#define TEXUNIT 0.99
#define PLACE(pos) (*(l_wellsdata + (l_vertwall1 + l_verteachwall * (i - 2)) * 5 + pos))
#define PLACE1(pos) (*(l_wellsdata + pos))
  l_verteachwall = 14 + (l_ndivwells - 1) * 2 * 4;
  
  l_wellsdata = malloc (VERTICETOT * sizeof (*l_wellsdata));
  
  if (!l_wellsdata)
  {
    perror ("Cannot allocate memory for wells vertices, aborting...");
    exit (EXIT_FAILURE);
  }
  else
  {
    /*
     *  When 2 values on the same line : texture coordinate 
     *  When 3 values on the same line : vertex
     */
    
    /* Tree initialization */
    PLACE1(0) = 1.;     PLACE1(1) = 0.;
    PLACE1(2) = HALF;   PLACE1(3) = 0.;         PLACE1(4) = 0.;
    PLACE1(5) = 0.;     PLACE1(6) = 0.;
    PLACE1(7) = -HALF;  PLACE1(8) = 0.;         PLACE1(9) = 0.;
    PLACE1(10) = 0.;    PLACE1(11) = 1.;
    PLACE1(12) = -HALF; PLACE1(13) = T_H * 2.;  PLACE1(14) = 0.;
    PLACE1(15) = 1.;    PLACE1(16) = 1.;
    PLACE1(17) = HALF;  PLACE1(18) = T_H * 2.;  PLACE1(19) = 0.;
    PLACE1(20) = 1.;    PLACE1(21) = 0.;
    PLACE1(22) = 0.;    PLACE1(23) = 0.;        PLACE1(24) = -HALF;
    PLACE1(25) = 0.;    PLACE1(26) = 0.;
    PLACE1(27) = 0.;    PLACE1(28) = 0.;        PLACE1(29) = HALF;
    PLACE1(30) = 0.;    PLACE1(31) = 1.;
    PLACE1(32) = 0.;    PLACE1(33) = T_H * 2.;  PLACE1(34) = HALF;
    PLACE1(35) = 1.;    PLACE1(36) = 1.;
    PLACE1(37) = 0.;    PLACE1(38) = T_H * 2.;  PLACE1(39) = -HALF;

    /* Loop on each wall to initialize */
    for (i = 2; i <= WALLLONG; i++)
    {
      int shift = 0; 
      PLACE(0)  = 0.;     PLACE(1) = TEXUNIT;
      PLACE(2)  = HALF;   PLACE(3) = T_H;       PLACE(4) = 0.;
      PLACE(5)  = 0.;     PLACE(6) = 0.;
      PLACE(7)  = HALF;   PLACE(8) = 0.;        PLACE(9) = 0.;

      /* loop to make a quarter of the wall round */
      for (j = 0; j < (l_ndivwells - 1); j++)
      {
        PLACE(10+j*10) = 1. / l_ndivwells * (j+1);                  PLACE(11+j*10) = TEXUNIT;
        PLACE(12+j*10) = cos (M_PI/(2*l_ndivwells)*(j+1)) * HALF;   PLACE(13+j*10) = T_H;       PLACE(14+j*10) = -sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
        PLACE(15+j*10) = 1. / l_ndivwells * (j+1);                  PLACE(16+j*10) = 0.;
        PLACE(17+j*10) = cos (M_PI/(2*l_ndivwells)*(j+1)) * HALF;   PLACE(18+j*10) = 0.;        PLACE(19+j*10) = -sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
      }

      /* set a value to be the shift number of vertices that the round part needed */
      shift = (l_ndivwells - 1) * 10;
      
      PLACE(10+shift) = 1.;                 PLACE(11+shift) = TEXUNIT;
      PLACE(12+shift) = 0.;                 PLACE(13+shift) = T_H;        PLACE(14+shift) = -HALF;
      PLACE(15+shift) = 1.;                 PLACE(16+shift) = 0.;
      PLACE(17+shift) = 0.;                 PLACE(18+shift) = 0.;         PLACE(19+shift) = -HALF; 
      PLACE(20+shift) = 1+i;                PLACE(21+shift) = TEXUNIT;
      PLACE(22+shift) = -T_C * (i - 1.);    PLACE(23+shift) = T_H;        PLACE(24+shift) = -HALF;
      PLACE(25+shift) = 1+i;                PLACE(26+shift) = 0.;
      PLACE(27+shift) = -T_C * (i - 1.);    PLACE(28+shift) = 0.;         PLACE(29+shift) = -HALF;

      /* Same loop on another quarter */
      for (j = 0; j < (l_ndivwells - 1); j++)
      {
        PLACE(30+shift+j*10) = 1+i + 1. / l_ndivwells * (j+1);                            PLACE(31+shift+j*10) = TEXUNIT;
        PLACE(32+shift+j*10) = -sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF - T_C * (i - 1.);  PLACE(33+shift+j*10) = T_H;     PLACE(34+shift+j*10) = -cos(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
        PLACE(35+shift+j*10) = 1+i + 1. / l_ndivwells * (j+1);                            PLACE(36+shift+j*10) = 0.;
        PLACE(37+shift+j*10) = -sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF - T_C * (i - 1.);  PLACE(38+shift+j*10) = 0.;      PLACE(39+shift+j*10) = -cos(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
      }
      
      shift = (l_ndivwells - 1) * 10 * 2;
      
      PLACE(30+shift)  = 1+i+1;             PLACE(31+shift) = TEXUNIT;
      PLACE(32+shift)  = -T_C * (i - 0.5);  PLACE(33+shift) = T_H;      PLACE(34+shift) = 0.;
      PLACE(35+shift)  = 1+i+1;             PLACE(36+shift) = 0.;
      PLACE(37+shift)  = -T_C * (i - 0.5);  PLACE(38+shift) = 0.;       PLACE(39+shift) = 0.;
      /*continue;*/
      for (j = 0; j < (l_ndivwells - 1); j++)
      {
        PLACE(40+shift+j*10) = 1+i+1 + 1. / l_ndivwells * (j+1);                          PLACE(41+shift+j*10) = TEXUNIT;
        PLACE(42+shift+j*10) = -cos(M_PI/(2*l_ndivwells)*(j+1)) * HALF - T_C * (i - 1.);  PLACE(43+shift+j*10) = T_H;     PLACE(44+shift+j*10) = sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
        PLACE(45+shift+j*10) = 1+i+1 + 1. / l_ndivwells * (j+1);                          PLACE(46+shift+j*10) = 0.;
        PLACE(47+shift+j*10) = -cos(M_PI/(2*l_ndivwells)*(j+1)) * HALF - T_C * (i - 1.);  PLACE(48+shift+j*10) = 0.;      PLACE(49+shift+j*10) = sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
      }
      
      shift = (l_ndivwells - 1) * 10 * 3;
      
      PLACE(40+shift) = 1+i+2;            PLACE(41+shift) = TEXUNIT;
      PLACE(42+shift) = -T_C * (i - 1.);  PLACE(43+shift) = T_H;      PLACE(44+shift) = HALF;
      PLACE(45+shift) = 1+i+2;            PLACE(46+shift) = 0.;
      PLACE(47+shift) = -T_C * (i - 1.);  PLACE(48+shift) = 0.;       PLACE(49+shift) = HALF;
      PLACE(50+shift) = 1+i+2+i;          PLACE(51+shift) = TEXUNIT;
      PLACE(52+shift) = 0.;               PLACE(53+shift) = T_H;      PLACE(54+shift) = HALF;
      PLACE(55+shift) = 1+i+2+i;          PLACE(56+shift) = 0.;
      PLACE(57+shift) = 0.;               PLACE(58+shift) = 0.;       PLACE(59+shift) = HALF;
      
      for (j = 0; j < (l_ndivwells - 1); j++)
      {
        PLACE(60+shift+j*10) = 1+i+2+i + 1. / l_ndivwells * (j+1);      PLACE(61+shift+j*10) = TEXUNIT;
        PLACE(62+shift+j*10) = sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF;  PLACE(63+shift+j*10) = T_H;   PLACE(64+shift+j*10) = cos(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
        PLACE(65+shift+j*10) = 1+i+2+i + 1. / l_ndivwells * (j+1);      PLACE(66+shift+j*10) = 0.;
        PLACE(67+shift+j*10) = sin(M_PI/(2*l_ndivwells)*(j+1)) * HALF;  PLACE(68+shift+j*10) = 0.;    PLACE(69+shift+j*10) = cos(M_PI/(2*l_ndivwells)*(j+1)) * HALF;
      }
      
      shift = (l_ndivwells - 1) * 10 * 4;
      
      PLACE(60+shift)  = 1+i+2+i+1;   PLACE(61+shift) = TEXUNIT;
      PLACE(62+shift)  = HALF;        PLACE(63+shift) = T_H;    PLACE(64+shift) = 0.;
      PLACE(65+shift)  = 1+i+2+i+1;   PLACE(66+shift) = 0.;
      PLACE(67+shift)  = HALF;        PLACE(68+shift) = 0.;     PLACE(69+shift) = 0.;
/* cubic wells not interesting for round ends
* verteachwall = 10 here
      PLACE(0)  = 0.; PLACE(1) = 1.;
      PLACE(2)  = HALF; PLACE(3) = T_H;  PLACE(4) = -HALF;
      PLACE(5)  = 0.; PLACE(6) = 0.;
      PLACE(7)  = HALF; PLACE(8) = 0.;   PLACE(9) = -HALF;
      PLACE(10) = i;  PLACE(11) = 1.;
      PLACE(12) = -T_C * (i - 0.5); PLACE(13) = T_H;  PLACE(14) = -HALF;
      PLACE(15) = i;  PLACE(16) = 0.;
      PLACE(17) = -T_C * (i - 0.5); PLACE(18) = 0.;  PLACE(19) = -HALF;
      PLACE(20) = i+1; PLACE(21) = 1.;
      PLACE(22) = -T_C * (i - 0.5); PLACE(23) = T_H;  PLACE(24) = HALF;
      PLACE(25) = i+1; PLACE(26) = 0.;
      PLACE(27) = -T_C * (i - 0.5); PLACE(28) = 0.;  PLACE(29) = HALF;
      PLACE(30) = i+1+i; PLACE(31) = 1.;
      PLACE(32) = HALF; PLACE(33) = T_H;  PLACE(34) = HALF;
      PLACE(35) = i+1+i; PLACE(36) = 0.;
      PLACE(37) = HALF; PLACE(38) = 0.;  PLACE(39) = HALF;
      PLACE(40) = i+1+i+1; PLACE(41) = 1.;
      PLACE(42) = HALF; PLACE(43) = T_H;  PLACE(44) = -HALF;
      PLACE(45) = i+1+i+1; PLACE(46) = 0.;
      PLACE(47) = HALF; PLACE(48) = 0.;  PLACE(49) = -HALF;    */
    }
  }
  
#undef HALF
#undef TEXUNIT
#undef PLACE
#undef PLACE1
}

/* create the lists to call for the environment */
void
drawworld_createTerrainLists (void)
{
  int number;
/* That is the value to give to draw a unit of texture otherwise is may repeat like 1.1 */
#define TEXUNIT 0.98

  glNewList (WALLMAX, GL_COMPILE);
  glBindTexture (GL_TEXTURE_2D, draw_getTexID(TEXID_EXT));
  glBegin (GL_QUAD_STRIP);
  glTexCoord2f (T_Y + T_X, 0.0);
  glVertex3f (+T_C / 2., 0.0, -T_C / 2.);
  glTexCoord2f (T_Y + T_X, TEXUNIT);
  glVertex3f (+T_C / 2., T_H * 2., -T_C / 2.);
  glTexCoord2f (T_Y + T_X / 2., 0.0);
  glVertex3f (-T_C * T_X + T_C / 2., 0.0, -T_C / 2.);
  glTexCoord2f (T_Y + T_X / 2., TEXUNIT);
  glVertex3f (-T_C * T_X + T_C / 2., T_H * 2., -T_C / 2.);
  glTexCoord2f (T_Y / 2. + T_X / 2., 0.0);
  glVertex3f (-T_C * T_X + T_C / 2., 0.0, -T_C / 2. + T_Y * T_C);
  glTexCoord2f (T_Y / 2. + T_X / 2., TEXUNIT);
  glVertex3f (-T_C * T_X + T_C / 2., T_H * 2., -T_C / 2. + T_Y * T_C);
  glTexCoord2f (T_Y / 2., 0.0);
  glVertex3f (+T_C / 2., 0.0, -T_C / 2. + T_Y * T_C);
  glTexCoord2f (T_Y / 2., TEXUNIT);
  glVertex3f (+T_C / 2., T_H * 2., -T_C / 2. + T_Y * T_C);
  glTexCoord2f (0.0, 0.0);
  glVertex3f (+T_C / 2., 0.0, -T_C / 2.);
  glTexCoord2f (0.0, TEXUNIT);
  glVertex3f (+T_C / 2., T_H * 2., -T_C / 2);
  glEnd ();
  glEndList ();

#undef TEXUNIT

  glNewList (TOPBOT, GL_COMPILE);

  glShadeModel (GL_SMOOTH);
  glBegin (GL_QUADS);
  glColor4f (l_ceilingColor[0], l_ceilingColor[1], l_ceilingColor[2],
             l_ceilingColor[3]);
  glVertex3f (+T_C / 2.0, T_H * 2., -T_C / 2.0);
  glColor4f (l_ceilingColor[0], l_ceilingColor[1], l_ceilingColor[2],
             l_ceilingColor[3]);
  glVertex3f (+T_C / 2.0, T_H * 2., T_Y * T_C - T_C / 2.0);
  glColor4f (l_ceilingColor[0], l_ceilingColor[1], l_ceilingColor[2],
             l_ceilingColor[3]);
  glVertex3f (-T_X * T_C + T_C / 2.0, T_H * 2., T_Y * T_C - T_C / 2.0);
  glColor4f (l_ceilingColor[0], l_ceilingColor[1], l_ceilingColor[2],
             l_ceilingColor[3]);
  glVertex3f (-T_X * T_C + T_C / 2.0, T_H * 2., -T_C / 2.0);
  glEnd ();
  glShadeModel (GL_FLAT);

  glEnable (GL_TEXTURE_2D);
  glBindTexture (GL_TEXTURE_2D, draw_getTexID(TEXID_GROUND));

  glBegin (GL_QUADS);
  glTexCoord2f (0.0, 0.0);
  glVertex3f (+T_C / 2.0, 0.0, -T_C / 2.0);
  glTexCoord2f (0.0, 16.0);
  glVertex3f (-T_X * T_C + T_C / 2.0, 0.0, -T_C / 2.0);
  glTexCoord2f (16.0, 16.0);
  glVertex3f (-T_X * T_C + T_C / 2.0, 0.0, T_Y * T_C - T_C / 2.0);
  glTexCoord2f (16, 0.0);
  glVertex3f (+T_C / 2.0, 0.0, T_Y * T_C - T_C / 2.0);
  glEnd ();

  glDisable (GL_TEXTURE_2D);

  glEndList ();

  glNewList (CLOUDS, GL_COMPILE);

  glEnable (GL_BLEND);
  glBindTexture (GL_TEXTURE_2D, draw_getTexID(TEXID_CLOUD));
  for (number = 30; number > 0; number--)
  {
    glPushMatrix ();
    glTranslatef (-(MYRAND - 0.25) * T_C * T_X * 2.,
                  number * 100 + T_H * 4.,
                  (MYRAND - 0.25) * T_C * T_Y * 2.);
    glRotatef (180, 1, 0, 0);
    glRotatef (MYRAND * 360., 0, 1, 0);
    glScalef ((MYRAND + 0.1) * 3., (MYRAND + 0.1) * 3.,
              (MYRAND + 0.1) * 3.);
    glBegin (GL_QUADS);
    glTexCoord2f (0.0, 0.0);
    glVertex3f (+T_C, 0.0, -T_C);
    glTexCoord2f (0.0, 1.0);
    glVertex3f (-T_C, 0.0, -T_C);
    glTexCoord2f (1.0, 1.0);
    glVertex3f (-T_C, 0.0, +T_C);
    glTexCoord2f (1, 0.0);
    glVertex3f (+T_C, 0.0, +T_C);
    glEnd ();
    glPopMatrix ();
  }
  glDisable (GL_BLEND);

  glEndList ();
}


/* set all what must be to launch the game -- graphic part */
void
drawworld_makeAll (void)
{
  drawworld_makeLobsterPincer ();
  drawworld_makeLobsterArms ();
  drawworld_makeLobsterAnt ();
  drawworld_makeLobsterTail ();
  drawworld_makeLobsterBody ();
  drawworld_makeLobsterLegs ();

  drawworld_makeMartianHead ();
  drawworld_makeMartianSoccer ();
}

/* Make available the vertex arrays */
void
drawworld_activateArray (void)
{
  glVertexPointer (3, GL_FLOAT, sizeof (GLfloat) * 5, l_wellsdata + 2);
  glTexCoordPointer (2, GL_FLOAT, sizeof (GLfloat) * 5, l_wellsdata);
  
#if USEGLEW && defined (GL_ARB_vertex_buffer_object)
  if (GLEW_ARB_vertex_buffer_object)
  {
    if (l_vbobuf)
      glDeleteBuffersARB (1, &l_vbobuf);
    glGenBuffersARB (1, &l_vbobuf);
    glBindBufferARB (GL_ARRAY_BUFFER, l_vbobuf);
    glBufferDataARB (GL_ARRAY_BUFFER, VERTICETOT * sizeof *l_wellsdata, NULL,
                     GL_STATIC_DRAW);
    glBufferSubDataARB (GL_ARRAY_BUFFER, 0, VERTICETOT * sizeof *l_wellsdata,
                        l_wellsdata);

#  define BUFFER_OFFSET(a) ((char *)NULL + (a))
    glVertexPointer (3, GL_FLOAT, sizeof (GLfloat) * 5, BUFFER_OFFSET (2 * sizeof (GLfloat)));
    glTexCoordPointer (2, GL_FLOAT, sizeof (GLfloat) * 5,
                       BUFFER_OFFSET (0));
#  undef BUFFER_OFFSET
  }
#endif /* USEGLEW && defined (GL_ARB_vertex_buffer_object) */

  glEnableClientState (GL_VERTEX_ARRAY);
  glEnableClientState (GL_TEXTURE_COORD_ARRAY);
}

/* create the list to call for the mini map */
void
drawworld_createMinimap (void)
{
  int length, number;

  glNewList (SMALLMAP, GL_COMPILE);

  glColor4f (1.0, 1.0, 1.0, 0.3);
  glRecti (0, 0, SM_PITCH * T_X, SM_PITCH * T_Y);

  glColor4f (0.0, 0.0, 0.0, 0.3);

/* loop on each wall */
  for (length = WALLLONG - 1; length >= 0; length--)
  {
    for (number = 0; number < NEWALL; number++)
    {
      switch (l_wells[length][number][2])
      {
        case 1:
          glRecti (l_wells[length][number][0] * SM_PITCH,
                   l_wells[length][number][1] * SM_PITCH,
                   (l_wells[length][number][0] + 1 + length) * SM_PITCH,
                   (l_wells[length][number][1] + 1) * SM_PITCH);
        break;
        case 2:
          glRecti (l_wells[length][number][0] * SM_PITCH,
                   l_wells[length][number][1] * SM_PITCH,
                   (l_wells[length][number][0] + 1) * SM_PITCH,
                   (l_wells[length][number][1] + 1 + length) * SM_PITCH);
        break;
        case 3:
          glRecti ((l_wells[length][number][0] - length) * SM_PITCH,
                   l_wells[length][number][1] * SM_PITCH,
                   (l_wells[length][number][0] + 1) * SM_PITCH,
                   (l_wells[length][number][1] + 1) * SM_PITCH);
        break;
        case 4:
          glRecti (l_wells[length][number][0] * SM_PITCH,
                   (l_wells[length][number][1] - length) * SM_PITCH,
                   (l_wells[length][number][0] + 1) * SM_PITCH,
                   (l_wells[length][number][1] + 1) * SM_PITCH);
        break;
      }
    }
  }
  glEndList ();
}

/* Draw all the wells */
void
drawworld_wells (void)
{
  int length, number;
  int dx = 0;
  int dy = 0;

  /* loop on each wall */
  for (length = WALLLONG - 1; length >= 0; length--)
  {
    for (number = 0; number < NEWALL; number++)
    {
      dx = dy = 0;
      switch (l_wells[length][number][2])
      {
        case 1:
          dx = length;
        break;
        case 2:
          dy = length;
        break;
        case 3:
          dx = -length;
        break;
        case 4:
          dy = -length;
        break;
      }
      if (!draw_isVisible (l_wells[length][number][0],    l_wells[length][number][1],    T_C/2.) &&
          !draw_isVisible (l_wells[length][number][0]+dx, l_wells[length][number][1]+dy, T_C/2.))
        continue; /* wall is in the back of the player */

      glPushMatrix ();
      glTranslatef (-l_wells[length][number][0] * T_C, 0.,
                    l_wells[length][number][1] * T_C);
      glRotatef ((l_wells[length][number][2] - 1) * 90., 0.0, 1.0, 0.0);
      glBindTexture (GL_TEXTURE_2D, draw_getTexID(TEXID_WALL1 + length));
      if (length == 0)
          glDrawArrays (GL_QUADS, 0, l_vertwall1);
      else
          glDrawArrays (GL_QUAD_STRIP, (length - 1) * l_verteachwall + l_vertwall1, l_verteachwall);
      glPopMatrix ();
    }
  }
}
