/*
 *  
 *  Copyright © 2008 Colomban "Ban" Wendling <ban@herbesfolles.org>
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 */

/*
 * Creation date: Fri Aug  1 16:09:05 2008
 * Revision date: Mon Sep  1 00:22:48 CEST 2008
 */


/* includes */
#include <stdio.h>
#include <stdlib.h>
#include "utf8.h"


/* FIXME: (partially done)
 * avoid invalid UTF-8 characters that have a valid value, e.g.
 * 11000000 10101110
 * that have the value of a dot (00101110 = 46 = '.') but that could be stored
 * with only one octet, here 00101110.
 * But avoiding empty first octet is not a way, because
 * 11100000 10111111 10111111
 * has a void first octet, but there's no shorter way to describe it because of 
 * the UTF-8 control bits.
 * 
 * (yes, with two octets, we should do:
 *  vvv      vv       // UTF-8 control bits
 *  11000000 10000000 // UTF-8 control mask
 *  |
 *  00111111 00111111 // value to store
 *  but this override the third UTF-8 control bit.
 *  then we have to extend with 3 octets:
 *  vvvv     vv       vv       // UTF-8 control bits
 *  11100000 10000000 10000000 // UTF-8 control mask
 *  |
 *  00000000 00111111 00111111 // value to store
 *  and that's right, we get
 *  =
 *  11100000 10111111 10111111
 *  we havn't overriden UTF-8 control bits.
 *  
 *  but in the same way, the value itself is representable with 4 octets, no
 *  problems, but it's no valid even if it doesn't override anything
 *  vvvvv    vv       vv       vv       // UTF-8 control bits
 *  11110000 10000000 10000000 10000000 // UTF-8 control mask
 *  |
 *  00000000 00000000 00111111 00111111 // value to store, through 4 octets
 *  you see here that the second octet is useless.
 * )
 */

#define U8_MSGPFX "U8: "


/** 
 * @brief Gets the minimal number of bits needed to store a value
 * @param c   an integer value
 * @returns the minimal number of bits needed to represent the given value.
 * 
 * This function gets the number of bits needed to store an integer value.
 * I.e., 255 takes 8 bits and 256 takes 9 bits.
 * 
 * \note This function considers that 0 needs 1 bit to be stored.
 * 
 * @internal
 */
static size_t bitlen (unsigned long c)
{
  size_t len = 0;
  
  /* loop while length is not enought to store value */
  while (c >= (1UL << ++len));
  
  return len;
}

/** 
 * @brief Gets the number of octets beeded to stores some bits
 * @param bits  a number of bits
 * @returns the number of octets needed to store the given number of bits
 * 
 * This function gets the minimal number of octets needed to store a given
 * number of bits.
 * 
 * @internal
 */
static size_t bits2octet (size_t bits)
{
  return (bits + ((bits % 8) ? 8 : 0)) / 8;
}

/** 
 * @brief Gets the value of first octet of an UTF-8 character
 * @param chu8  an UTF-8 character
 * @param len   the length of the character
 * @returns the value part of the character
 * 
 * This function gets the value part of an UTF-8 character stored in its first
 * octet. This is NOT the value of the character but only witch one that is
 * stored in the first octet.
 * 
 * @warning this function works only for the first octet. Use u8_suboctetval()
 *          for the other octets.
 * @warning this function doesn't check the validity of the octet, you must gave
 *          a valid octet.
 * 
 * @see u8_suboctetval()
 * 
 * @internal
 */
static unsigned long u8_firstoctetval (const char* chu8,
                                       size_t len)
{
  long val;
  
  /* with single octet character, all the octet is the value */
  if (len == 1)
    val = (unsigned long)*chu8;
  /* else, remove UTF-8 control bits to get the value */
  else
    val = (unsigned long)
          /* remove character size mask, e.g. 1110xxxx for a 3-bytes one */
          ((((unsigned char)chu8[0]) ^
           /* get character's size mask */
           (255 >> (8-len) << (8-len)))
          /* then move it to its offset (nbytes -1).
           * 6 is because subcharacter is coded in 6 bits, the 2 MSB are used
           * to design in-character bytes */
          << (6*(len -1)));
  
  return val;
}

/** 
 * @brief Gets the value of an UTF-8 subcharacter octet
 * @param chu8    an UTF-8 character
 * @param len     the length of the character
 * @param offest  the offset of the octet in the character.
 * @returns the value part of the character
 * 
 * This function gets the value part of an UTF-8 character stored in one of its
 * suboctets. This is NOT the value of the character but only witch one that is
 * stored in the octet at @p offset.
 * 
 * @warning this function doesn't work for the first octet. Use
 *          u8_firstoctetval() instead.
 * @warning this function doesn't check the validity of the octet, you must gave
 *          a valid octet.
 * 
 * @see u8_firstoctetval()
 * 
 * @internal
 */
static unsigned long u8_suboctetval (const char* chu8,
                                     size_t len,
                                     size_t offset)
{
  return (unsigned long)
         /* remove 10xxxxxx (flag of in-character bytes) */
         (((unsigned char)chu8[offset]) ^ 128)
         /* then move it to its position (see u8_firstoctetval ()) */
         << (6*(len -1 -offset));
}

/** 
 * @brief Checks if a value is valid for a length
 * @param val   the value stored in a character
 * @param len   the length of the character
 * @return 1 if the character is valid, 0 otherwise
 * 
 * This function cheks if a value is valid for a charcter of @p len octets.
 * This is useful because UTF-8 character must be shorter as possible.
 */
/* FIXME: verify this function */
/* FIXME: do a real check, ideally compatible with all UTF-8-like octets, not
 * only the UTF-8 one; e.g. with 8-octets long characters */
int u8_valisvalid (long val, size_t len)
{
  int valid = 0;
  
  /* the values are mostly 1<<((6*n)+(8-(2+n))) except for 0x80 and 0x0 */
  if ((len == 1 && val < 0x80 && val >= 0x0) ||
      (len == 2 && val < 0x800 && val >= 0x80) ||
      (len == 3 && val < 0x10000 && val >= 0x800) ||
      (len == 4 && val < 0x200000 && val >= 0x10000) /* ||
      (len == 5 && val < 0x4000000 && val >= 0x200000) ||
      (len == 6 && val < 0x80000000 && val >= 0x4000000) ||
      (len == 7 && val < 0x1000000000 && val >= 0x80000000) || */)
    valid = 1;
  
  return valid;
}

/** 
 * @brief Gets the length in bytes of an UTF-8 character
 * @param chu8  an UTF-8 character
 * @returns the octet length of @p chu8 or 0 if the character is wrong
 * 
 * This function gets the length os an UTF-8 character.
 * 
 * @note If you pass an UTF-8 string as the character, only the first character
 *       will be computed.
 * 
 * @warning this function does not check if the character is really valid, but
 *          just if its UTF-8 control bits are.
 */
size_t u8_chlen (const char* chu8)
{
  int i;
  size_t len = 0;
  
  /* count the true MSBs to get the number of octets of the character, i.e. 3
   * for a 1110xxxx octet */
  for (i=7; (unsigned char)chu8[0] & (1 << i) && i >= 0; i--)
    len++;
  
  /* single-byte character (0: single-byte character has no control bits) */
  if (len == 0)
    len = 1;
  /* multibyte character (UTF-8 does not allow chracters with more than 4
   * octets) */
  else if (len >= 2 && len <= 4)
  {
    /* we want a size_t type for compatibility with @v len */
    size_t j;
    
    /* check in-character bytes validity */
    for (j=1; j < len; j++)
    {
      /* check if the byte follows the convention 10xxxxxx */
      if (!((unsigned char)chu8[j] & 128) || (unsigned char)chu8[j] & 64)
      {
        len = 0;
#ifdef U8_DEBUG
        fprintf (stderr, U8_MSGPFX"invalid UTF-8 character: invalid byte value '%.3u' at offset %lu\n", (unsigned char)chu8[j], j);
#endif
        break;
      }
    }
  }
  /* not an UTF-8 character */
  else
  {
#ifdef U8_DEBUG
    fprintf (stderr, U8_MSGPFX"invalid UTF-8 character: ");
    if (len == 1)
      fprintf (stderr, "unexpected in-character byte %u\n", (unsigned char)chu8[0]);
    else
      fprintf (stderr, "invalid byte value '%u' at offset 0\n", (unsigned char)chu8[0]);
#endif
    len = 0;
  }
  
  return len;
}

/** 
 * @brief Gets the value of an UTF-8 character
 * @param chu8  an UTF-8 character
 * @returns the Unicode value of the character or -1 if the character is wrong
 * 
 * @note This function checks if the value contained in the caharcter is valid
 *       for it. E.g. An overlong character is considered as invalid conforming to 
 */
/* we beed long because UTF-8 can store character through 21 bytes */
long u8_chval (const char* chu8)
{
  size_t len;
  size_t i;
  long val = -1L; /* default is invalid character/nul character */
  
  /* get the character length (in bytes) */
  len = u8_chlen (chu8);
  
  /* valid character */
  if (len >= 1)
  {
    val = u8_firstoctetval (chu8, len);
    
    /* multicotets characters */
    if (len > 1)
    {
      /* add each suoctets values */
      for (i=1; i < len; i++)
        val |= u8_suboctetval (chu8, len, i);
    }
    
    /* check if the value found is valid */
    if (!u8_valisvalid (val, len))
      val = -1L;
  }
  
  return val;
}

/** 
 * @brief Gets real offset of a character in an UTF-8 string
 * @param stru8   an UTF-8 string
 * @param offset  the offset in characters
 * @returns the offset in bytes or -1 it the string is invalid
 * 
 * This function is designed to be used as char array offset, instead of the C
 * standard offset, computing character length to get the real offset.
 * For example, in an UTF-8 string "éàa", the third character, here 'a' has an
 * offset of 4 decause both 'é' and 'à' are coded through two bytes.
 * You can use this function to correct an array offset, e.g.:
 *  char str_utf8[] = "éèa€";
 *  str_utf8[1] = 'a'; // wrong, gets the second byte of 'é'
 *  str_utf8[u8_choff(str_utf8, 1)] = 'a' // right
 * This is a bad example, because we replace a multibytes character by a
 * single-byte one, and this should break the string. But the idea is here.
 */
long u8_choff (const char* stru8, long offset)
{
  long i, j;
  size_t chlen;
  
  for (i=0, j=0; j < offset; j++)
  {
    chlen = u8_chlen (&stru8[i]);
    if (chlen == 0)
    {
#ifdef U8_DEBUG
      fprintf (stderr, U8_MSGPFX"invalid UTF-8 character at %lu\n", i);
#endif
      return -1L;
    }
    
    i += chlen;
  }
  
  return i;
}

/**
 * @brief Gets the length in characters of an UTF-8 string
 * @param stru8   a NUL-ended UTF-8 string
 * @returns the length of @p stru8 in characters or -1 if the string is invalid
 * 
 * This function gets the number of UTF-8 characters found in the given string.
 * The string must be a C-like string, ending by a NUL character.
 */
/* long type is because this function can return -1 */
long u8_strlen (const char* stru8)
{
  long len = 0;
  size_t chlen;
  
  while (*stru8 != 0)
  {
    chlen = u8_chlen (stru8);
    if (chlen == 0)
    {
#ifdef U8_DEBUG
      fprintf (stderr, U8_MSGPFX"invalid UTF-8 character at %lu\n", len);
#endif
      return -1L;
    }
    
    len ++;
    
    stru8 += chlen;
  }
  
  return len;
}

/** 
 * @brief Creates an UTF-8 character from its unicode value
 * @param val   an Unicode value
 * @returns a newl allocated UTF-8 character that you should free after use or
 *          NULL on memory allocatrion failure.
 * 
 * This function creates an UTF-8 character from its Unicode code.
 * 
 * @warning this function does not check the validity of the given value, then
 *          it can crate an UTF-8 representing a wrong Unicode character,
 *          or even try to create a wrong UTF-8 character.
 */
/* FIXME: this function don't check the validity of the result, it is just
 * computed as UTF-8 can be; but a bad unicode (too big) value can be computed
 * too
 */
char* u8_val2ch (unsigned long val)
{
  unsigned char* chu8 = NULL;
  
  /* single byte character */
  if (val < 128)
  {
    chu8 = malloc (1);
    if (chu8 == NULL)
      fprintf (stderr, U8_MSGPFX"failed to allocate 1 byte of memory\n");
    else
      *chu8 = (unsigned char)val;
  }
  /* multibytes character */
  else
  {
    size_t bits;
    size_t bytes;
    size_t tmpbit;
    int i;
    size_t j;
    
    /* bits for the character itself (value part, without UTF-8 control bits) */
    bits = bitlen (val);
    bytes = bits2octet (bits);
    /* then compute the UTF-8 bit length */
    bits += (2*(bytes-1)) + bytes + 1;
    /* ...and bytes length */
    bytes = bits2octet (bits);
    
    chu8 = malloc (bytes);
    if (chu8 == NULL)
      fprintf (stderr, U8_MSGPFX"failed to allocate %lu bytes of memory\n", bytes);
    else
    {
      /* process first byte */
      /* set base bitmask of the first byte (e.g. 1110xxxx for a 3-bytes char) */
      chu8[0] = 0;
      for (i=0; (size_t)i < bytes; i++)
        chu8[0] |= 1 << (7 - i);
      /* then set the character part value for this byte */
      for (i = (6 - bytes); i >= 0; i--)
      {
        /* bit to test */
        tmpbit = (1 << (6*(bytes-1))) << i;
        
        if (val & tmpbit)
        {
          /* add the bit to the character */
          chu8[0] |= tmpbit >> (6*(bytes-1));
          /* remove it from the value */
          val ^= tmpbit;
        }
      }
      
      /* process next bytes */
      for (j = 1; j < bytes; j++)
      {
        /* default bitmask (10xxxxxx) */
        chu8[j] = 128;
        
        /* almost the same as above */
        for (i=5; i >= 0; i--)
        {
          tmpbit = (1 << (6*(bytes-j-1))) << i;
          
          if (val & tmpbit)
          {
            chu8[j] |= tmpbit >> (6*(bytes-j-1));
            val ^= tmpbit;
          }
        }
      }
    }
  }
  
  return (char*)chu8;
}

